const koaStatic = require('koa-static');
const { defaultsDeep } = require('lodash/fp');
const jwt = require('jsonwebtoken');
const fs = require('fs');
const path_1 = require('path');
const logging = require('@strapi/logger');
const { isNull } = require('util');
const utils = require('@strapi/utils');
const { ApplicationError, UnauthorizedError, ForbiddenError, NotFoundError } =
  utils.errors;

const logger = logging.createLogger();

const defaults = {
  maxAge: 60000,
  defaultIndex: true,
};

const getOwnerIdPath = async () => {
  try {
    const configOwnerIdPath = await strapi
      .documents('api::config.config')
      .findOne({
        fields: ['ownerIdPath'],
      });
    return configOwnerIdPath.ownerIdPath
      ? configOwnerIdPath.ownerIdPath.split('.')
      : [];
  } catch {
    return [];
  }
};

const buildPopulateObject = (arr) => {
  if (arr.length === 0) return null;

  let populateObject = {};
  let currentLevel = populateObject;

  arr.forEach((key, index) => {
    currentLevel[key] = {};
    if (index < arr.length - 1) {
      currentLevel[key].populate = {};
      currentLevel = currentLevel[key].populate;
    }
  });

  return populateObject;
};

const getUserDynamicPopulate = async (userId, populateArray) => {
  const populateObject = buildPopulateObject(populateArray);

  const queryOptions = populateObject ? { populate: populateObject } : {};

  queryOptions.documentId = userId;

  return await strapi
    .documents('plugin::users-permissions.user')
    .findOne(queryOptions);
};

const getOwnerId = (user, ownerIdPath) => {
  if (ownerIdPath.length === 0) return user.id;

  let currentLevel = user;

  ownerIdPath.forEach((key, index) => {
    currentLevel = currentLevel[key];
  });

  return currentLevel.id;
};

const getUserId = (ctx, jwtToken) => {
  const authHeader = ctx.request.header.authorization;
  let decoded;
  try {
    if (authHeader.startsWith('Bearer ')) {
      const token = authHeader.substring(7, authHeader.length);
      decoded = jwt.verify(token, jwtToken);
    } else {
      throw new Error('Token invalid or missing');
    }
  } catch (error) {
    logger.warn('Token invalid or missing');
    throw new UnauthorizedError('Token invalid or missing', {});
  }
  return decoded.id;
};

module.exports = (plugin, config) => {
  const rootPath = path_1.resolve(strapi.dirs.static.public, '..');
  const SUPER_USER = process.env.SUPER_USER
    ? process.env.SUPER_USER
    : 'i-taros-user';
  const OWNER_ID_FIELD_NAME = process.env.OWNER_ID_FIELD_NAME
    ? process.env.OWNER_ID_FIELD_NAME
    : 'ZuwendungsempfaengerId';
  console.log('super-user', SUPER_USER);
  const { defaultIndex, maxAge } = defaultsDeep(defaults, config);
  const jwtToken = process.env.JWT_SECRET;

  strapi.server.routes([
    {
      method: 'GET',
      path: '/private/(.*)',
      handler: koaStatic(rootPath, {
        maxage: maxAge,
        defer: true,
      }),
      config: {
        auth: false,
        policies: [
          async (policyContext, config, { strapi }) => {
            logger.info('Checking private route Policy');

            const userId = getUserId(policyContext, jwtToken);

            const ownerIdPath = await getOwnerIdPath();

            let userOwnerId, user;
            if (ownerIdPath && ownerIdPath.length > 0) {
              user = await getUserDynamicPopulate(userId, ownerIdPath);
              if (!user) {
                throw new ApplicationError('User does not exist');
              }
              if (user.username === SUPER_USER) {
                logger.info('User has permission to read the private folder');
                return true;
              }
              try {
                userOwnerId = getOwnerId(user, ownerIdPath);
              } catch (error) {
                throw new ApplicationError(
                  'OwnerIdPath config setting invalid',
                  {}
                );
              }
            } else {
              user = await strapi.db
                .query('plugin::users-permissions.user')
                .findOne({
                  where: { id: userId },
                });
              if (!user) {
                throw new ApplicationError('User does not exist');
              }
              userOwnerId = user[OWNER_ID_FIELD_NAME];
            }

            // path = "/private/1/year/{filename}"
            const ownerIdFolder = policyContext.request.url.split('/')[3];
            if (
              (!isNaN(ownerIdFolder) && userOwnerId == ownerIdFolder) ||
              user.username === SUPER_USER
            ) {
              logger.info('User has permission to read the private folder');
              return true;
            } else {
              logger.warn('User has no permission to read the private folder.');
              throw new ForbiddenError(
                'User has no permission to read the private folder',
                {}
              );
            }
          },
        ],
      },
    },
    {
      method: 'GET',
      path: '/public/(.*)',
      handler: koaStatic(rootPath, {
        maxage: maxAge,
        defer: true,
      }),
      config: {
        auth: false,
        policies: [
          async (policyContext, config, { strapi }) => {
            // TODO
            // Does user has to be authenticated to read public files?
            logger.info('in GET /public Policy Check: ');
          },
        ],
      },
    },
    {
      method: 'POST',
      path: '/api/upload',
      handler: koaStatic(strapi.dirs.static.public, {
        maxage: maxAge,
        defer: true,
      }),
      config: {
        auth: false,
        policies: [
          async (policyContext, config, { strapi }) => {
            logger.info('Checking POST route Policy');

            const userId = getUserId(policyContext, jwtToken);

            if (policyContext.request.body.type == 'public') {
              return true;
            } else if (policyContext.request.body.type == 'private') {
              const ownerIdPath = await getOwnerIdPath();

              let userOwnerId, user;
              if (ownerIdPath && ownerIdPath.length > 0) {
                user = await getUserDynamicPopulate(userId, ownerIdPath);
                if (!user) {
                  throw new ApplicationError('User does not exist');
                }
                if (user.username === SUPER_USER) {
                  logger.info(
                    'User has permission to write to the private folder'
                  );
                  return true;
                }
                try {
                  userOwnerId = getOwnerId(user, ownerIdPath);
                } catch (error) {
                  throw new ApplicationError(
                    'OwnerIdPath config setting invalid',
                    {}
                  );
                }
              } else {
                user = await strapi.db
                  .query('plugin::users-permissions.user')
                  .findOne({
                    where: { id: userId },
                  });
                if (!user) {
                  throw new ApplicationError('User does not exist');
                }
                userOwnerId = user[OWNER_ID_FIELD_NAME];
              }
              // path = "/private/1/year/{filename}"
              const ownerIdFolder = policyContext.request.body.ownerId;

              if (
                (!isNaN(ownerIdFolder) && userOwnerId == ownerIdFolder) ||
                user.username === SUPER_USER
              ) {
                logger.info(
                  'User has permission to write to the private folder'
                );
                return true;
              } else {
                logger.warn(
                  'User has no permission to write to the private folder.'
                );
                throw new ForbiddenError(
                  'User has no permission to write the private folder',
                  {}
                );
              }
            } else {
              throw new ApplicationError(
                "Required request part 'type' for method parameter type String must be 'public' or 'private'",
                {}
              );
            }
          },
        ],
      },
    },
    {
      method: 'DELETE',
      path: '/api/upload/files/(.*)',
      handler: koaStatic(rootPath, {
        maxage: maxAge,
        defer: true,
      }),
      config: {
        auth: false,
        policies: [
          async (policyContext, config, { strapi }) => {
            logger.info('Checking private route Policy');

            const fileId = policyContext.request.url.split('/')[4];
            const file = await strapi.plugins['upload'].services.upload.findOne(
              fileId
            );

            if (file == null) {
              throw new NotFoundError('file.notFound', {});
            }

            // path = "/private/1/year/{filename}"
            const ownerIdFolder = file.url.split('/')[3];
            const fileType = file.url.split('/')[1];
            if (fileType == 'public') {
              return true;
            } else {
              const userId = getUserId(policyContext, jwtToken);
              const ownerIdPath = await getOwnerIdPath();

              let userOwnerId, user;
              if (ownerIdPath && ownerIdPath.length > 0) {
                user = await getUserDynamicPopulate(userId, ownerIdPath);
                if (!user) {
                  throw new ApplicationError('User does not exist');
                }
                if (user.username === SUPER_USER) {
                  logger.info(
                    'User has permission to delete file from the private folder'
                  );
                  return true;
                }
                try {
                  userOwnerId = getOwnerId(user, ownerIdPath);
                } catch (error) {
                  throw new ApplicationError(
                    'OwnerIdPath config setting invalid',
                    {}
                  );
                }
              } else {
                user = await strapi.db
                  .query('plugin::users-permissions.user')
                  .findOne({
                    where: { id: userId },
                  });
                if (!user) {
                  throw new ApplicationError('User does not exist');
                }
                userOwnerId = user[OWNER_ID_FIELD_NAME];
              }
              if (
                (!isNaN(ownerIdFolder) && userOwnerId == ownerIdFolder) ||
                user.username === SUPER_USER
              ) {
                logger.info(
                  'User has permission to delete file from the private folder'
                );
                return true;
              } else {
                logger.warn(
                  'User has no permission to delete file from the private folder.'
                );
                throw new ForbiddenError(
                  'User has no permission to delete file from the private folder',
                  {}
                );
              }
            }
          },
        ],
      },
    },
    {
      method: 'GET',
      path: '/api/upload/files/(.*)',
      handler: koaStatic(rootPath, {
        maxage: maxAge,
        defer: true,
      }),
      config: {
        auth: false,
        policies: [
          async (policyContext, config, { strapi }) => {
            logger.info('Checking private route Policy');

            const fileId = policyContext.request.url.split('/')[4];
            const file = await strapi.plugins['upload'].services.upload.findOne(
              fileId
            );

            if (file == null) {
              throw new NotFoundError('file.notFound', {});
            }
            const fileType = file.url.split('/')[1];
            if (fileType == 'public') {
              return true;
            }

            const userId = getUserId(policyContext, jwtToken);

            // get ownerIdPath from config content-type as array
            const ownerIdPath = await getOwnerIdPath();

            let userOwnerId, user;
            if (ownerIdPath && ownerIdPath.length > 0) {
              user = await getUserDynamicPopulate(userId, ownerIdPath);
              if (!user) {
                throw new ApplicationError('User does not exist');
              }
              if (user.username === SUPER_USER) {
                logger.info(
                  'User has permission to read file from the private folder'
                );
                return true;
              }
              try {
                userOwnerId = getOwnerId(user, ownerIdPath);
              } catch (error) {
                throw new ApplicationError(
                  'OwnerIdPath config setting invalid',
                  {}
                );
              }
            } else {
              user = await strapi.db
                .query('plugin::users-permissions.user')
                .findOne({
                  where: { id: userId },
                });
              if (!user) {
                throw new ApplicationError('User does not exist');
              }
              userOwnerId = user[OWNER_ID_FIELD_NAME];
            }
            // path = "/private/1/year/{filename}"
            const ownerIdFolder = file.url.split('/')[3];

            if (
              (!isNaN(ownerIdFolder) && userOwnerId == ownerIdFolder) ||
              user.username === SUPER_USER
            ) {
              logger.info(
                'User has permission to read file from the private folder'
              );
              return true;
            } else {
              logger.warn(
                'User has no permission to read file from the private folder.'
              );
              throw new ForbiddenError(
                'User has no permission to read file from the private folder',
                {}
              );
            }
          },
        ],
      },
    },
  ]);
  return plugin;
};
